#
# program to turn a quake1/2/3 map into a `planes' file, which has planes for
# all brushes of the map (of worldspawn class)
# (note that the format is slightly different in doom3 .map files)
#
# Alex S.
#

use strict;

# check
die "specify .map file, and output file as parameters" if $#ARGV < 1;
my $file = $ARGV[0];
my $outfile = $ARGV[1];

# load file
open my $in,$file or die $!;
read $in,my $cont,-s $file or die $!;
close $in;

# get rid of comments.
$cont =~ s|\/\/.*?\n||sgi;

# replace '{' and '}' with properly nested <tagN>...</tagN>, where N is the depth.
my $cnt = 0;
$cont =~ s/(\{|\})/$1 eq '{' ? '<tag'.(++$cnt).'>':'<\/tag'.($cnt--).'>'/sgie;

my $ents = []; # list of ents.
my %ents = (); # ents by name (classname)

# loop for all entities.
while($cont =~ m|<tag1>(.+?)</tag1>|sgi){
    my $entity = trim($1);
    my $ent = {prim=>[],vals=>{}};
    # read all of that entities brushes/patches.
    while($entity =~ m|<tag2>(.*?)</tag2>|sgi){
        my $primitive = trim($1);
        if($primitive !~ m|patch|si){  # is this a brush? (ignore patches)
            $primitive =~ s{.*<tag3>|</tag3>.*}{}sgi;
            my $brush = {type=>'brush', sides => []};
            for my $face (split /\s*\n\s*/,trim($primitive)){
                my @p1 = ($1,$2,$3)
                    if $face =~ s|\(\s+(\S+)\s+(\S+)\s+(\S+)\s+\)\s+||;
                my @p2 = ($1,$2,$3)
                    if $face =~ s|\(\s+(\S+)\s+(\S+)\s+(\S+)\s+\)\s+||;
                my @p3 = ($1,$2,$3)
                    if $face =~ s|\(\s+(\S+)\s+(\S+)\s+(\S+)\s+\)\s+||;
                my $material = $1 if $face =~ s|(\S+?)||;
                my @plane = vec2plane(@p2,@p1,@p3);
                # ignore the rest for now.
                push @{$brush->{sides}},{plane=>\@plane,material=>$material};
            }
            push @{$ent->{prim}},$brush;
        }
    }
    # erase all of brushes/patches
    $entity =~ s|<tag2>.*</tag2>||sgi;

    # what's left is name-value pairs for that entity; read those.
    while($entity =~ m|\"(.*?)\"\s+\"(.*?)\"|sgi){
        $ent->{vals}{$1} = $2;
    }
    push @{$ents},$ent;
    $ents{$ent->{vals}{classname}} = $ent;
}

# cut off beginning and trailing spaces
sub trim {
    my ($s) = @_;
    $s =~ s/^\s+|\s+$//sg;
    return $s;
}

# some vector ops

sub vecsub {
    my ($ax,$ay,$az,$bx,$by,$bz) = @_;
    return ($ax - $bx,$ay - $by,$az - $bz);
}

sub veccross {
    my ($ax,$ay,$az,$bx,$by,$bz) = @_;
    return ($ay*$bz-$az*$by,$az*$bx-$ax*$bz,$ax*$by-$ay*$bx);
}

sub vecdot {
    my ($ax,$ay,$az,$bx,$by,$bz) = @_;
    return $ax*$bx+$ay*$by+$az*$bz;
}

sub vecnorm {
    my ($x,$y,$z) = @_;
    my $len = sqrt(vecdot($x,$y,$z,$x,$y,$z));
    return ($x/$len,$y/$len,$z/$len);
}

# Plane is: Ax + By + Cx + D = 0
sub vec2plane {
    my @d1 = vecsub(@_[3..5],@_[0..2]);
    my @d2 = vecsub(@_[6..8],@_[0..2]);
    my @crs = veccross(@d1,@d2);
    my @norm = vecnorm(@crs);
    my $d = -vecdot(@_[0..2],@norm);
    return (@norm,$d);
}

open my $out,'>'.$outfile or die $!;

# only write out worldspawn
my $ent = $ents{worldspawn};
for my $prim (@{$ent->{prim}}[0..299]){   # change this to support many more brushes
    next if $prim->{type} ne 'brush';
    print $out ($#{$prim->{sides}}+1)." ";
    for my $s (@{$prim->{sides}}){
        print $out "$s->{plane}[0] $s->{plane}[1] $s->{plane}[2] $s->{plane}[3] ";
    }
}
close $out;
